"""
Zoom API Helper - Utilities to interact with the Zoom API v2
"""

__title__ = 'zoom-api-helper'
__description__ = 'Utilities to interact with the Zoom API v2'
__url__ = 'https://github.com/rnag/zoom-api-helper'
__version__ = '0.1.1'
__author__ = 'Ritvik Nag'
__author_email__ = 'rv.kvetch@gmail.com'
__license__ = 'MIT'
__copyright__ = 'Copyright 2022 Ritvik Nag'



from __future__ import annotations

__all__ = [
    'cached_property',
    'CustomEncoder',
    'log_time',
    'read_json_file_if_exists',
    'save_json_file',
    'write_to_csv',
]

from csv import DictWriter
from datetime import datetime
from json import load, dump, JSONEncoder
from time import time
from typing import TYPE_CHECKING, Any, TypeVar

from functools import wraps

try:
    from functools import cached_property
except ImportError:
    # noinspection PyUnresolvedReferences, PyPackageRequirements
    from backports.cached_property import cached_property

from .log import LOG


class CustomEncoder(JSONEncoder):

    def default(self, o: Any) -> Any:
        if isinstance(o, datetime):
            return o.isoformat()

        return JSONEncoder.default(self, o)


if TYPE_CHECKING:
    _F = TypeVar('_F')


def log_time(func: _F) -> _F:

    @wraps(func)
    def inner(*args, _fn_name=func.__name__, **kwargs):
        start = time()
        ret = func(*args, **kwargs)
        end = time()

        LOG.debug('[%s] completed in %.3fs', _fn_name, end - start)

        return ret

    return inner


def read_json_file_if_exists(filename: str) -> dict | list | None:
    try:
        with open(filename) as in_file:
            return load(in_file)

    except FileNotFoundError:
        return None


def save_json_file(filename: str, data: dict | list):
    with open(filename, 'w') as out_file:
        dump(data, out_file)


def write_to_csv(out_file: str, rows: list[dict]):
    with open(out_file, 'w') as output:
        writer = DictWriter(output, fieldnames=rows[0].keys())
        writer.writeheader()
        writer.writerows(rows)



__all__ = [
    'Meeting',
]

from enum import Enum
from typing import TYPE_CHECKING


# noinspection GrazieInspection
if TYPE_CHECKING:
    __all__ += ['RowType', 'ProcessRow', 'UpdateRow']

    from datetime import datetime
    from typing import Any

    try:
        from typing import Protocol
    except ImportError:  # Python 3.7
        from typing_extensions import Protocol

    # Type of Excel Row.
    #
    # Typically, `sheet2dict` passes this in as `dict[str, str]`, however the
    # type restrictions can be lifted a little, so that a user assignment
    # won't result in warnings.
    RowType = dict[str, str | datetime | bool | int | float | dict | list | None]


    class ProcessRow(Protocol):

        def __call__(self, row: RowType) -> bool:
            ...


    class UpdateRow(Protocol):

        def __call__(self, row: RowType, resp: dict[str, Any]) -> None:
            ...


class Meeting(Enum):
    """
    The type of Zoom meeting.
    """
    INSTANT = 1
    SCHEDULED = 2
    RECURRING = 3
    RECURRING_WITH_TIME = 8



"""Project-specific constant values."""

__all__ = [
    'API_USERS',
    'CACHE_DIR',
    'CREATE_MEETING_KWARGS',
]

from os import getenv
from pathlib import Path


API_USERS = 'https://api.zoom.us/v2/users'

# Cache Directory
CACHE_DIR = Path(getenv('CACHE_DIR', '~/.zoom/cache')).expanduser()

# Create Cache Directory if needed.
CACHE_DIR.mkdir(parents=True, exist_ok=True)

# Create Meeting API - valid keyword arguments
#
# Ref: https://marketplace.zoom.us/docs/api-reference/zoom-api/methods/#operation/meetingCreate
CREATE_MEETING_KWARGS = {
    # Extras: used by our `create_meeting` method
    'host_id', 'host_email',
    # Zoom API keyword arguments
    'agenda', 'start_time', 'template_id', 'password', 'timezone', 'topic',
    'tracking_fields', 'duration', 'recurrence', 'default_password',
    'pre_schedule', 'settings', 'schedule_for', 'type'
}



from __future__ import annotations

import json
import os
from concurrent.futures import ThreadPoolExecutor, as_completed
from datetime import datetime
from pathlib import Path
from typing import Any

import requests

from .constants import *
from .log import LOG
from .models import *
from .oauth import get_access_token
from .utils import *


class ZoomAPI:
    # noinspection GrazieInspection
    """
    Helper client to interact with the `Zoom API v2`_

    .. _Zoom API v2: https://marketplace.zoom.us/docs/api-reference/zoom-api/

    """

    __slots__ = (
        # required for `@cached_property`
        '__dict__',
        # instance attributes
        '_session',
        '_account_id',
        '_access_token',
        '_client_id',
    )

    def __init__(self, client_id: str,
                 client_secret: str,
                 account_id: str = None,
                 local=False):

        self._session = session = requests.Session()
        self._account_id = account_id = account_id or os.getenv('ZOOM_ACCOUNT_ID')
        self._client_id = client_id

        if local:
            self._access_token = token = 'abc12345'
        else:
            self._access_token = token = get_access_token(
                session, account_id, client_id, client_secret,
            )

        session.headers['Authorization'] = f'Bearer {token}'
        session.headers['Content-Type'] = 'application/json'

    @classmethod
    def dummy_client(cls):
        """Create a dummy :class:`ZoomAPI` client, for testing purposes."""
        return cls('...', '...', local=True)

    def __repr__(self):
        cls_name = self.__class__.__name__
        masked_token = f'{self._access_token[:5]}***'
        return (f'{cls_name}(access_token={masked_token!r}, '
                f'account_id={self._account_id!r}, '
                f'client_id={self._client_id!r})')

    def _new_session(self) -> requests.Session:
        session = requests.Session()
        session.headers['Authorization'] = f'Bearer {self._access_token}'
        session.headers['Content-Type'] = 'application/json'

        return session

    def _get(self, url: str, params: dict = None, session: requests.Session | None = None):
        LOG.debug('GET %s, params=%s', url, params)
        return (session or self._session).get(url, params=params)

    def _post(self, url: str, data: dict = None, session: requests.Session | None = None):
        LOG.debug('POST %s, data=%s', url, data)
        return (session or self._session).post(url, json=data)

    @log_time
    def bulk_create_meetings(self, col_header_to_kwarg: dict[str, str] = None,
                             *,
                             rows: list[RowType] = None,
                             excel_file: str | os.PathLike[str] = None,
                             max_threads=10,
                             process_row: ProcessRow = None,
                             update_row: UpdateRow = None,
                             out_file: str | os.PathLike[str] = None,
                             default_timezone='UTC',
                             dry_run=False):
        # noinspection GrazieInspection
        """``POST /users/{userId}/meetings``: Use this API to *bulk*-create meetings, given
        a list of meetings to create.

        If the rows containing meetings to create lives in an Excel (.xlsx) file,
        then `excel_file` must be passed in, and contain the filepath of the Excel
        file to retrieve the meeting details from; else, `rows` must be passed in
        with a list of meetings to create.

        Note that to read from Excel, the ``sheet2dict`` library is required;
        this can be installed easily via::

            $ pip install zoom-api-helper[excel]

        ``col_header_to_kwarg`` is a mapping of column headers to keyword arguments,
        as accepted by the :meth:`create_meeting` method. If the header names are
        all title-cased versions of the keyword arguments, then this argument does
        *not* need to be passed in. See also, :attr:`constants.CREATE_MEETING_KWARGS`
        for a full list of acceptable keywords for the *Create Meeting* API; note
        that these are specified as *values* in the key-value pairing.

        ``process_row`` is an optional function or callable that will be called with
        a copy of each *row*, or individual meeting info. The function can modify
        the row in place as desired.

        ``update_row`` is an optional function or callable that will be called with
        the HTTP response data from the Create Meetings API, and the associated
        row from the Excel file. The function can modify the row in place as desired.

        If ``dry_run`` is enabled, then no API calls will be made, and this function
        will essentially be a no-op; however, useful logging output is printed for
        debugging purposes. Enable this parameter along with the :func:`setup_logging`
        helper function, in a debug environment.

        **How It Works**

        This function scans in a list of rows containing a list of meetings
        to create, either from a local file or from the ``rows`` parameter.
        Then, it calls ``process_row`` on each row, and also retrieves a
        mapping of column name to keyword argument via ``col_header_to_kwarg``.
        Then, it concurrently creates a list of meetings via the Zoom
        **Create Meetings** API. Finally, it calls ``update_row`` on each
        response and row pair, and writes out the updated meeting info to
        an output CSV file named ``out_file``, which defaults to
        ``{excel-file-name-without-ext}}.out.csv`` if an output filepath
        is not specified.

        **References**

        API documentation:
            - https://marketplace.zoom.us/docs/api-reference/zoom-api/methods/#operation/meetingCreate

        If a naive `datetime` object is provided for ``start_time`` or any other field
        (i.e. one with no timezone information) the value for ``timezone`` will determine
        which timezone to associate with the `datetime` object - defaults to *UTC* time
        if not specified.

        For a list of supported timezones, please see:
            - https://marketplace.zoom.us/docs/api-reference/other-references/abbreviation-lists/#timezones

        """
        # noinspection PyUnresolvedReferences
        import sheet2dict

        def to_snake_case(s: str):
            return s.replace(' ', '_').replace('-', '_').lower()

        if rows:
            col_headers = rows[0].keys()
        else:
            ws = sheet2dict.Worksheet()
            ws.xlsx_to_dict(excel_file)
            col_headers = ws.sheet_items[0].keys()
            rows = ws.sheet_items

        header_to_kwarg = {kwarg: kwarg for kwarg in CREATE_MEETING_KWARGS}

        if not col_header_to_kwarg:
            col_header_to_kwarg = {h: to_snake_case(h) for h in col_headers}

        for h in col_headers:
            kwarg = col_header_to_kwarg.get(h) or to_snake_case(h)
            if kwarg in CREATE_MEETING_KWARGS:
                header_to_kwarg[h] = kwarg

        meetings_to_create = []

        if process_row is None:
            def do_nothing(_row): return True
            process_row = do_nothing

        for row in rows:
            # copy row so as not to modify it directly.
            copied_row = row.copy()

            # optional: process the row.
            is_valid = process_row(copied_row)
            if not is_valid:
                continue

            # retrieve meeting info.
            mtg = {header_to_kwarg[h]: copied_row[h] for h in header_to_kwarg
                   if h in copied_row}

            # add default fields.
            if 'timezone' not in mtg:
                mtg['timezone'] = default_timezone

            meetings_to_create.append(mtg)

        # if it's a dry run, print useful info for debugging purposes, then quit.
        if dry_run:
            print('Column Header to Keyword Argument:')
            print(json.dumps(header_to_kwarg, indent=2))
            print()
            print(f'Have {len(meetings_to_create)} Meetings to Create:')
            print(json.dumps(meetings_to_create, indent=2, cls=CustomEncoder))
            return

        LOG.debug('Column Header to Keyword Argument: %s', header_to_kwarg)

        if out_file is None:
            p = Path(excel_file)
            out_file = p.with_name(f'{p.stem}.out.csv')

        LOG.debug('Output File: %s', out_file.absolute())

        def create_meeting(mtg_: RowType):
            # use a separate session for each thread.
            session = self._new_session()
            return self.create_meeting(session=session, **mtg_)

        if update_row is None:
            update_row = dict.update

        with ThreadPoolExecutor(max_workers=max_threads) as executor:
            # Start the create meeting operations and mark each future with its row index
            future_to_row_idx = {executor.submit(create_meeting, mtg): i
                                 for i, mtg in enumerate(meetings_to_create)}

            for future in as_completed(future_to_row_idx):
                idx = future_to_row_idx[future]
                row = rows[idx]

                try:
                    resp = future.result()
                except Exception as exc:
                    LOG.error('[%d] %r generated an exception: %s', idx, row, exc)
                else:
                    update_row(row, resp)

        write_to_csv(out_file, rows)

    def create_meeting(self, *,
                       session: requests.Session | None = None,
                       host_id: str | None = None,
                       host_email: str | None = None,
                       topic: str = 'My Meeting',
                       agenda: str = 'My Description',
                       start_time: datetime | str | None = None,
                       timezone: str | None = 'UTC',
                       type: Meeting | None = None,
                       **request_data) -> dict[str, Any]:
        """``POST /users/{userId}/meetings``: Use this API to create a meeting for a user.

        Ref:
            - https://marketplace.zoom.us/docs/api-reference/zoom-api/methods/#operation/meetingCreate

        If a naive `datetime` object is provided for `start_time` or any other field
        (i.e. one with no timezone information) the value for`timezone` will determine
        which timezone to associate with the `datetime` object - defaults to *UTC* time
        if not specified.

        For a list of supported timezones, please see:
            - https://marketplace.zoom.us/docs/api-reference/other-references/abbreviation-lists/#timezones

        :return:
        """
        if host_id:
            user_id = host_id
        elif host_email:
            user_id = self._user_email_to_id_cached[host_email]
        else:
            user_id = 'me'

        if agenda:
            request_data['agenda'] = agenda

        if topic:
            request_data['topic'] = topic

        if start_time:
            request_data['start_time'] = start_time.isoformat() \
                if isinstance(start_time, datetime) else start_time

        if timezone:
            request_data['timezone'] = timezone

        if type:
            request_data['type'] = type.value

        r = self._post(f'https://api.zoom.us/v2/users/{user_id}/meetings', request_data, session)
        r.raise_for_status()

        return r.json()

    @cached_property
    def _user_email_to_id_cached(self):
        return self.user_email_to_id(use_cache=True)

    def user_email_to_id(self, status: str | None = 'active', *, use_cache=False):
        filename = CACHE_DIR / f'users_{self._account_id}_{self._client_id}.json'

        if use_cache:
            users = read_json_file_if_exists(filename)
            if users:
                return users

        users = self.list_users(status)['users']
        email_to_id = {u['email']: u['id'] for u in users}

        # save list of users to cache
        save_json_file(filename, email_to_id)

        return email_to_id

    def list_users(self, status: str | None = 'active', page_size=300, page=1, all_pages=True):
        # noinspection GrazieInspection
        """``GET /users``: Use this API to list your account's users.

        Ref:
            - https://marketplace.zoom.us/docs/api-reference/zoom-api/methods/#operation/users

        :param status: The user's status, one of: active, inactive, pending
        :param page_size: The number of records returned within a single API call.
        :param page: The page number of the current page in the returned records.
        :param all_pages: True to paginate and retrieve all records (pages).
        :return: A `dict` object, where the `users` key contains a list of users.

        """
        params = {'page_size': page_size, 'page_number': page}
        if status:
            params['status'] = status

        res = self._get(API_USERS, params=params)
        final_data = data = res.json()

        if all_pages:
            while data['page_count'] > data['page_number']:
                params['page_number'] += 1
                res = self._get(API_USERS, params=params)
                data = res.json()
                final_data['users'].extend(data['users'])

        return final_data



"""
Zoom API Helper
~~~~~~~~~~~~~~~

Utilities to interact with the Zoom API v2

Sample Usage:

    >>> import zoom_api_helper

For full documentation and more advanced usage, please see
<https://zoom-api-helper.readthedocs.io>.

:copyright: (c) 2022 by Ritvik Nag.
:license: MIT, see LICENSE for more details.
"""

__all__ = [
    'ZoomAPI',
    'setup_logging',
]


from .v2 import ZoomAPI
from .log import setup_logging



from __future__ import annotations

from logging import NullHandler, basicConfig, getLogger, DEBUG, WARNING
from typing import TYPE_CHECKING


LOG = getLogger('zoom_api_helper')

# Set up logging to ``/dev/null`` like a library is supposed to.
# http://docs.python.org/3.3/howto/logging.html#configuring-logging-for-a-library
LOG.addHandler(NullHandler())

if TYPE_CHECKING:
    LevelType = str | int


def setup_logging(default_level: LevelType = DEBUG,
                  lib_level: LevelType = DEBUG,
                  urllib3_level: LevelType = WARNING):
    """
    Sets up logging for application (user) code.
    """
    basicConfig(level=default_level)

    LOG.setLevel(lib_level)
    getLogger('urllib3').setLevel(urllib3_level)



"""Tests for `zoom_api_helper` package."""

import pytest


from zoom_api_helper import ZoomAPI


def test_create_zoom_client():
    """Sample pytest test function with the pytest fixture as an argument."""
    zoom = ZoomAPI.dummy_client()
    print(zoom)



"""The setup script."""
import itertools
import pathlib

from setuptools import setup, find_packages


here = pathlib.Path(__file__).parent

package_name = 'zoom_api_helper'

packages = find_packages(include=[package_name, f'{package_name}.*'])

requires = [
    'requests',
    'backports.cached-property~=1.0.2; python_version == "3.7"',
    'typing-extensions; python_version == "3.7"',
]

test_requirements = [
    'pytest~=7.1.3',
    'pytest-cov~=3.0.0',
]

extras_require = {
   'excel': ['sheet2dict'],
}

# Ref: https://stackoverflow.com/a/71166228/10237506
extras_require['all'] = list(itertools.chain.from_iterable(extras_require.values()))


about = {}
exec((here / package_name / '__version__.py').read_text(), about)

readme = (here / 'README.rst').read_text()
history = (here / 'HISTORY.rst').read_text()

setup(
    name=about['__title__'],
    version=about['__version__'],
    description=about['__description__'],
    long_description=readme + '\n\n' + history,
    long_description_content_type='text/x-rst',
    author=about['__author__'],
    author_email=about['__author_email__'],
    url=about['__url__'],
    packages=packages,
    include_package_data=True,
    install_requires=requires,
    project_urls={
        'Documentation': 'https://zoom-api-helper.readthedocs.io',
        'Source': 'https://github.com/rnag/zoom-api-helper',
    },
    license=about['__license__'],
    # TODO add more relevant keywords as needed
    keywords=['zoom-api-helper', 'zoom', 'api', 'zoom-api', 'api-v2'],
    classifiers=[
        # Ref: https://pypi.org/classifiers/
        'Development Status :: 5 - Production/Stable',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: MIT License',
        'Natural Language :: English',
        'Programming Language :: Python :: 3 :: Only',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
        'Programming Language :: Python :: 3.10',
        'Programming Language :: Python'
],
    test_suite='tests',
    extras_require=extras_require,
    tests_require=test_requirements,
    zip_safe=False
)



from time import time

import requests

from .constants import CACHE_DIR
from .utils import read_json_file_if_exists, save_json_file


def get_access_token(session: requests.Session,
                     account_id: str,
                     client_id: str,
                     client_secret: str) -> str:
    """Retrieve an access token, given credentials for a Server-to-Server OAuth app.

    To use account credentials to get an access token for your app, call the Zoom OAuth token API
    with the account_credentials `grant_type` and your `account_id`.

    The successful response will be the access token, which is a Bearer token type that expires
    in an hour, with the scopes that you chose in your app settings screen.

    Ref:
        - https://marketplace.zoom.us/docs/guides/build/server-to-server-oauth-app/


    :param session: (Optional) Requests Session
    :param account_id: Zoom Account ID
    :param client_id: OAuth app Client ID
    :param client_secret: OAuth app Client Secret

    :return: The access token for the app.
    """
    filename = CACHE_DIR / f'token_{account_id}_{client_id}.json'

    # first, check the cache for the access token.
    cache = read_json_file_if_exists(filename)
    if cache and cache['expires_at'] > round(time()):
        return cache['access_token']

    # else, we make a live API call to retrieve the token.
    params = {
        'grant_type': 'account_credentials',
        'account_id': account_id,
    }

    r = session.post(
        'https://zoom.us/oauth/token',
        auth=(client_id, client_secret),
        params=params
    )
    r.raise_for_status()

    data = r.json()

    token = data['access_token']
    # 5 minutes before the token actually expires
    expires_at = round(time()) + data['expires_in'] - (5 * 60)

    # save the access token to the cache.
    cache = {'access_token': token, 'expires_at': expires_at}
    save_json_file(filename, cache)

    return token
